using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

using Atoms;
using NTree;
using NParsing;
using NDecoding;

/* TODO :
 * High priority ::
 *        1] simplify expressions smartly by analyzing parenthesis/priorities/ "y / y^2" ..etc
 * Not so high priority ::
 *        2] instead of "variables disappearing", allow (force!) d/dx(x+y) to be evaluated using (x,y)
 *        3] maybe distinguish operators and functions, although they are similar
 *        4] allow the constants e, pi to be used in expressions
 */
namespace NExpression
{
    interface IExpression
    {
        string ToString();

        double Eval(params double[] args);

        Expression Diff(char variable);
    }

    public class Expression : IExpression 
    {
        public string strExp { get; set;} // string representing the expression
        private List<IAtom> listExp; // stack representing the expression
        private TreeAtom treeExp; // tree representing the expression
        private bool formatted;
        private bool parsed;

        // build an expression from user input
        public Expression(string str)
        {
            strExp = string.Copy(str);
            formatted = false;
            parsed = false;
        }

        private void ParseAll()
        {
            if (parsed) return;
            if (!formatted) strExp = Parse.Format(strExp);
            formatted = true;
            listExp = Parse.ToRPN(strExp);
            treeExp = Parse.ToTree(listExp);
            parsed = true;
        }

        public override string ToString()
        {
            if (!formatted) strExp = Parse.Format(strExp); // to ensure string is formatted
            formatted = true;
            return string.Copy(strExp);
        }

        // evaluate the expression for a set of parameters
        // CONVENTION: params. given in alphabetical order ; "a f u x y" for ex.
        public double Eval(params double[] args)
        {
            ParseAll(); // in case of..
            var listExpEval = Substitute(listExp, args);
            var values = new Stack<double>();

            // use stack representation to substitute iteratively
            foreach (var elem in listExpEval)
            {
                if ((elem as IAtom).GetAType() == AType.Val) values.Push((elem as Value).Content);
                else
                {
                    // an operation must be realized
                    Function f = elem as Function;
                    double[] svals = new double[f.Arity];
                    for (int i = f.Arity - 1; i >= 0; i--)
                    {
                        svals[i] = values.Pop();
                    }
                    values.Push(f.Eval(svals));
                }
            }

            return values.Pop();
        }

        // realize the substitution, to get the computation ready
        private static List<IAtom> Substitute(List<IAtom> latom, params double[] args) 
        {
            // presentVars contains all present variables (names)
            var presentVars = new List<char>();
            // array of variables names per term ('_' if no variable, constant term)
            var varNames = new char[latom.Count];
            for (int i = 0; i < latom.Count; i++) varNames[i] = '_';

            int counter = 0;
            foreach (var elem in latom)
            {
                var a = elem as IAtom;
                if (a.GetAType() == AType.Var)
                {
                    char c = (a as Variable).Name;
                    presentVars.Add(c);
                    varNames[counter] = c;
                }
                counter++;
            }

            presentVars = presentVars.Distinct().ToList(); // removes duplicate variables
            presentVars.Sort();
            var varsIndex = new int[26]; // indices of (present) variables in args array
            counter = 0;
            foreach (var v in presentVars)
            {
                varsIndex[v - 'a'] = counter++;
            }

            // last pass: substitutes values to each variable
            var latomValues = new List<IAtom>();
            counter = 0;
            foreach (var elem in latom)
            {
                if (varNames[counter] != '_')
                {
                    latomValues.Add(new Value(args[varsIndex[varNames[counter] - 'a']]));
                }
                else latomValues.Add(elem as IAtom);
                counter++;
            }
            return latomValues;
        }

        // Differentiates the expression according to the selected variable.
        // Rather complex task:
        // 1) expression->list->tree
        // 2) recursive calls to get the derivated tree (easy, the root gives the rule)
        // 3) "decode" : resulting tree->list->expression
        public Expression Diff(char variable)
        {
            ParseAll(); // in case of
            TreeAtom tatom = treeExp.Diff(variable);
            int height = TreeAtom.GetHeight(tatom);
            for (int i = 0; i < height; i++) // upper bound number of calls, to ensure the most simplifications
            {
                if (tatom == null) break;
                tatom = tatom.Simplify(); // to avoid ugly "0.0 - 0.0", and so on..
            }
            if (tatom == null) tatom = new TreeAtom(new Value(0.0));
            var latom = Decode.Tree2List(tatom);
            return Decode.List2Exp(latom);
        }
    }

}
