using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;

using Atoms;
using NTree;

namespace NParsing
{
    public class Parse {

        // return true if c is part of a value (dot or digit ;
        // exponents [eg. "1e10"] conflict with variables, so unhandled for the moment)
        private static bool IsInValue(char c)
        {
            return ( (c >= '0' && c <= '9') || c == '.');
        }

        private static bool IsVariable(char c)
        {
            return (c >= 'a' && c <= 'z');
        }

        // transform the expression string into a clean "standard" form
        public static string Format(string strToParse)
        {
            strToParse = strToParse.ToLower();
            strToParse = strToParse.Replace('[', '(');
            strToParse = strToParse.Replace(']', ')');
            var strExp = "";

            bool lastInVal = false;
            bool lastIsVar = false; // to check for log, exp, and more generally functions names
            foreach (var c in strToParse)
            {
                // ignore consecutive spaces
                if (c == ' ') continue;
                
                // test if c is inside a value (number)
                bool currInVal = false;
                bool currIsVar = false;
                if (IsInValue(c)) currInVal = true;
                else if (IsVariable(c)) currIsVar = true;
                if ( (lastInVal && currInVal) || (lastIsVar && currIsVar) )
                {
                    // remove last inserted space:
                    strExp = strExp.Remove(strExp.Length - 1, 1);
                }
                lastInVal = currInVal;
                lastIsVar = currIsVar;
                strExp += (c + " ");
            }

            // finally, simplify the possible situations "x + - 3" and "x - - 3". 
            strExp = strExp.Replace("+ - ", "- ");
            strExp = strExp.Replace("- -", "+");
            return strExp;
        }

        private static bool IsFunction(string s)
        {
            return (!IsInValue(s[0]) &&
                (s.Length > 1 || s[0] == '+' || s[0] == '-' || s[0] == '*' || s[0] == '/' || s[0] == '^'));
        }

        // get a stack of atoms (operands, variables and values) representing the (infix) expression
        // in reversed polish notation
        public static List<IAtom> ToRPN(string strExp)
        {
            var listExp = new List<IAtom>();
            var staPO = new Stack<ParOrOp>(); // Parenthesis and Operators
            
            // tokenize string into substrings : values, variables, operators or parenthesis
            var tokens = strExp.Split(new char[] { ' ' });

            for (int i = 0; i < tokens.Length; i++)
            {
                if (tokens[i].Length == 0) break;
                if (IsInValue(tokens[i][0]))
                {
                    listExp.Add(new Value(Double.Parse(tokens[i], CultureInfo.InvariantCulture)));
                }
                else if (IsVariable(tokens[i][0])) // variables or functions names
                {
                    if (tokens[i].Length > 1)
                    {
                        // some function called by f(...,...,...)
                        staPO.Push(Function.CreateFunc(tokens[i]));
                    }
                    else // a variable
                    {
                        listExp.Add(new Variable(tokens[i][0]));
                    }
                }
                // special case 1: pow operator, highest priority
                else if (tokens[i][0] == '^')
                {
                    staPO.Push(Function.CreateFunc(tokens[i]));
                }
                else if (tokens[i][0] == '*' || tokens[i][0] == '/')
                {
                    // binary '*' and '/', high priority but lower than '^' and unary +, -
                    while (staPO.Count > 0)
                    {
                        // while the stack-top contains pow, uplus/uminus, unstack it and add to the RPN
                        var poa = staPO.Peek();
                        if (poa as Parenthesis == null &&
                            ((poa as Function).Name == "pow" || (poa as Function).Name == "uplus" || (poa as Function).Name == "uminus"))
                        {
                            listExp.Add(staPO.Pop() as Function);
                        }
                        else break;
                    }
                    staPO.Push(Function.CreateFunc(tokens[i]));
                }
                else if (tokens[i][0] == '+' || tokens[i][0] == '-')
                {
                    // special case 2: plus and minus, as they are entered same (by user) for unary and binary
                    if (i == 0 || tokens[i - 1][0] == '(' || IsFunction(tokens[i - 1]))
                    {
                        // unary + or - : high priority, but lower than '^'
                        while (staPO.Count > 0)
                        {
                            // while the stack-top contains an operator '^', unstack it and add to the RPN
                            ParOrOp poa = staPO.Peek();
                            if (poa as Parenthesis == null && (poa as Function).Name == "pow")
                            {
                                listExp.Add(staPO.Pop() as Function);
                            }
                            else break;
                        }
                        staPO.Push(Function.CreateFunc(tokens[i], 1));
                    }
                    else // binary operator of lower priority
                    {
                        while (staPO.Count > 0)
                        {
                            // while the stack-top contains an operator '^', unstack it and add to the RPN
                            ParOrOp poa = staPO.Peek();
                            if (poa as Parenthesis == null &&
                                ((poa as Function).Name == "pow" || (poa as Function).Name == "uplus" || (poa as Function).Name == "uminus"
                                || (poa as Function).Name == "mult" || (poa as Function).Name == "div"))
                            {
                                listExp.Add(staPO.Pop() as Function);
                            }
                            else break;
                        }
                        staPO.Push(Function.CreateFunc(tokens[i]));
                    }
                }
                else if (tokens[i][0] == ',') // separator for functions arguments
                {
                    while (staPO.Peek() as Parenthesis == null)
                    {
                        listExp.Add(staPO.Pop() as Function);
                    }
                }
                else if (tokens[i][0] == '(')
                {
                    staPO.Push(new Parenthesis('('));
                }
                else if (tokens[i][0] == ')')
                {
                    // unstack until left parenthesis :
                    while (staPO.Peek() as Parenthesis == null)
                    {
                        listExp.Add(staPO.Pop() as Function);
                    }
                    staPO.Pop(); // ..and unstack left parenthesis
                }
            }
            // unstack all remaining operators :
            while (staPO.Count > 0) listExp.Add(staPO.Pop() as Function);
            return listExp;
        }

        // get a tree of Values, Variables and Functions representing the expression
        public static TreeAtom ToTree(List<IAtom> listExp) {
            var trees = new Stack<TreeAtom>();

            IEnumerator lenum = listExp.GetEnumerator();
            while (lenum.MoveNext())
            {
                if (lenum.Current as Function == null) trees.Push(new TreeAtom(lenum.Current as IAtom));
                else
                {
                    // an operation must be realized (between trees !)
                    Function f = lenum.Current as Function;
                    TreeAtom[] ts = new TreeAtom[f.Arity];
                    for (int i=f.Arity-1; i>=0; i--) ts[i] = trees.Pop();
                    trees.Push(TreeAtom.Op(f.Name, ts));
                }
            }
            return trees.Pop();
        }
    }

}
