using System;

using Atoms;

namespace NTree
{
    // convention : if unary operator, only the most left part is not null
    public class TreeAtom
    {
        private static int MAX_ITER = 10;

        private static Random rGen = new Random();

        public IAtom Root { get; set; }

        public TreeAtom[] NextGen { get; set; }

        public TreeAtom()
        {
            NextGen = null;
        }

        public TreeAtom(IAtom r)
        {
            Root = r;
            NextGen = null;
        }

        public TreeAtom(IAtom r, params TreeAtom[] ts)
        {
            Root = r;
            NextGen = new TreeAtom[ts.Length];
            for (int i = 0; i < ts.Length; i++)
                NextGen[i] = ts[i];
        }

        public bool IsLeaf()
        {
            if (NextGen == null) return true;
            for (int i = 0; i < NextGen.Length; i++)
            {
                if (NextGen[i] != null) return false;
            }
            return true;
        }

        public static int GetHeight(TreeAtom t)
        {
            if (t == null) return 0;
            if (t.IsLeaf()) return 1;
            int H = 0;
            for (int i = 0; i < t.NextGen.Length; i++)
                H = Math.Max(H, GetHeight(t.NextGen[i]));
            return 1 + H;
        }

        // some_operation(T, S, ...)
        public static TreeAtom Op(string opName, params TreeAtom[] ts)
        {
            return new TreeAtom(Function.CreateFunc(opName), ts);
        }

        // main differentiation function
        public TreeAtom Diff(char variable)
        {
            return Diff_(this, variable);
        }

        // differentiate a tree recursively, according to a variable 'c'
        private static TreeAtom Diff_(TreeAtom t, char c)
        {
            if (t == null) return null;
            if (t.IsLeaf())
            {
                if (t.Root as Variable != null && (t.Root as Variable).Name == c)
                {
                    return new TreeAtom(new Value(1.0));
                }
                return new TreeAtom(new Value(0.0));
            }

            // switch according to the main operation == root:
            switch ((t.Root as Function).Name)
            {
                case "uplus":
                    return Diff_(t.NextGen[0], c);
                case "plus":
                    return Op("plus", Diff_(t.NextGen[0], c), Diff_(t.NextGen[1], c));
                case "uminus":
                    return Op("minus", Diff_(t.NextGen[0], c));
                case "minus":
                    return Op("minus", Diff_(t.NextGen[0], c), Diff_(t.NextGen[1], c));
                case "mult":
                    return Op("plus",
                        Op("mult", Diff_(t.NextGen[0], c), t.NextGen[1]),
                        Op("mult", t.NextGen[0], Diff_(t.NextGen[1], c))
                        );
                case "div":
                    return Op("div",
                        Op("minus",
                            Op("mult", Diff_(t.NextGen[0], c), t.NextGen[1]),
                            Op("mult", t.NextGen[0], Diff_(t.NextGen[1], c))
                            ),
                        Op("pow", t.NextGen[1], new TreeAtom(new Value(2.0)))
                        );
                case "pow":
                    return Op("mult",
                        Op("plus",
                            Op("mult", Op("log", t.NextGen[0]), Diff_(t.NextGen[1], c)),
                            Op("div", Op("mult", Diff_(t.NextGen[0], c), t.NextGen[1]), t.NextGen[0])
                            ),
                        Op("pow", t.NextGen[0], t.NextGen[1])
                        );
                case "exp":
                    return Op("mult", Diff_(t.NextGen[0], c), Op("exp", t.NextGen[0]));
                case "log":
                    return Op("div", Diff_(t.NextGen[0], c), t.NextGen[0]);
                case "sqrt":
                    return Op("mult",
                        new TreeAtom(new Value(0.5)), Op("pow", t.NextGen[0], new TreeAtom(new Value(-0.5)))
                        );
                case "cos":
                    return Op("uminus", Op("sin", t.NextGen[0]));
                case "sin":
                    return Op("cos", t.NextGen[0]);
                case "tan":
                    return Op("div", Op("sin", t.NextGen[0]), Op("cos", t.NextGen[1]));
                case "acos":
                    return Op("div",
                        new TreeAtom(new Value(-1.0)),
                        Op("sqrt", Op("minus",
                            new TreeAtom(new Value(1.0)),
                            Op("pow", t.NextGen[0], new TreeAtom(new Value(2.0)))
                            ))
                        );
                case "asin":
                    return Op("div",
                        new TreeAtom(new Value(1.0)),
                        Op("sqrt", Op("minus",
                            new TreeAtom(new Value(1.0)),
                            Op("pow", t.NextGen[0], new TreeAtom(new Value(2.0)))
                            ))
                        );
                case "atan":
                    return Op("div",
                        new TreeAtom(new Value(1.0)),
                        Op("plus",
                            new TreeAtom(new Value(1.0)),
                            Op("pow", t.NextGen[0], new TreeAtom(new Value(2.0)))
                            )
                        );
                case "cosh":
                    return Op("sinh", t.NextGen[0]);
                case "sinh":
                    return Op("cosh", t.NextGen[0]);
                case "tanh":
                    return Op("div",
                        new TreeAtom(new Value(1.0)),
                        Op("pow", Op("cosh", t.NextGen[0], new TreeAtom(new Value(2.0))))
                        );
            }
            return null; // never reached
        }

        // recursively delete null subtrees
        public TreeAtom Simplify()
        {
            if (IsLeaf())
            {
                return new TreeAtom(Root.Copy());
            }
            double[] tabRand = new double[26];
            double minTab = Double.MaxValue;
            double maxTab = Double.MinValue;
            for (int i = 0; i < MAX_ITER; i++)
            {
                // generate 26 random values (too many, but avoid long analysis)
                for (int j = 0; j < 26; j++) tabRand[j] = (2 * rGen.Next(2) - 1) * rGen.NextDouble();
                double tmpVal = Eval(tabRand);
                if (tmpVal < minTab) minTab = tmpVal;
                if (tmpVal > maxTab) maxTab = tmpVal;
            }
            if (minTab == maxTab)
            {
                // constant branches (maybe 0) are replaced by their values
                return new TreeAtom(new Value(minTab));
            }
            TreeAtom[] nexts = new TreeAtom[NextGen.Length];
            for (int i = 0; i < NextGen.Length; i++) nexts[i] = NextGen[i].Simplify();
            return Op((Root as Function).Name, nexts);
        }

        private double Eval(double[] tabRands)
        {
            // 26 random values --> eval expression recursively
            if (IsLeaf())
            {
                if (Root.GetAType() == AType.Val) return (Root as Value).Content;
                return tabRands[(Root as Variable).Name - 'a'];
            }
            double[] valsNext = new double[NextGen.Length];
            for (int i = 0; i < NextGen.Length; i++)
            {
                valsNext[i] = NextGen[i].Eval(tabRands);
            }
            return (Root as Function).Eval(valsNext);
        }
    }

}
